from fretboardgtr.elements.tuning import Tuning, TuningConfig


def test_tuning_get_svg():
    tuning = Tuning("test", position=(0.0, 0.0))
    attribs = tuning.get_svg().attribs
    assert float(attribs["x"]) == 0.0
    assert float(attribs["y"]) == 0.0


def test_tuning_get_svg_custom_config():
    tuning_config = TuningConfig(fontsize=30)
    tuning = Tuning("test", position=(0.0, 0.0), config=tuning_config)
    attribs = tuning.get_svg().attribs
    assert float(attribs["x"]) == 0.0
    assert float(attribs["y"]) == 0.0
    assert attribs["font-size"] == 30



from fretboardgtr.utils import (
    _contains_duplicates,
    chromatic_position_from_root,
    chromatics_from_root,
    note_to_interval,
    note_to_interval_name,
    scale_to_enharmonic,
    scale_to_flat,
    scale_to_intervals,
    scale_to_sharp,
    to_flat_note,
    to_sharp_note,
)


def test_contains_duplicates_false():
    elements = ["a", "b", "c"]
    assert _contains_duplicates(elements) is False


def test_contains_duplicates_true():
    elements = ["a", "b", "c", "a"]
    assert _contains_duplicates(elements) is True


def test_chromatics_from_root_a():
    root = "A"
    results = ["A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"]
    assert results == chromatics_from_root(root)


def test_chromatics_from_root_c():
    root = "C"
    results = [
        "C",
        "C#",
        "D",
        "D#",
        "E",
        "F",
        "F#",
        "G",
        "G#",
        "A",
        "A#",
        "B",
    ]
    assert results == chromatics_from_root(root)


def test_chromatics_position_from_root():
    assert 3 == chromatic_position_from_root(root="A", note="C")
    assert 7 == chromatic_position_from_root(root="C", note="G")


def test_to_sharp_note_a():
    assert "A" == to_sharp_note("A")


def test_to_sharp_note_a_sharp():
    assert "A#" == to_sharp_note("Bb")


def test_to_flat_note_a():
    assert "A" == to_flat_note("A")


def test_to_flat_note_a_sharp():
    assert "Bb" == to_flat_note("A#")


def test_scale_to_sharp():
    scale = ["A", "Bb", "B", "C", "C#"]
    assert ["A", "A#", "B", "C", "C#"] == scale_to_sharp(scale)


def test_scale_to_flat():
    scale = ["A", "Bb", "B", "C", "C#"]
    assert ["A", "Bb", "B", "C", "Db"] == scale_to_flat(scale)


def test_note_to_interval():
    root = "C"
    note = "G"
    assert 7 == note_to_interval(note, root)


def test_note_to_interval_name():
    root = "C"
    note = "G"
    assert "5" == note_to_interval_name(note, root)


def test_scale_to_intervals():
    scale = ["C", "E", "G"]
    assert [0, 4, 7] == scale_to_intervals(scale, root="C")


def test_scale_to_enharmonic():
    scale = ["G", "A", "B", "C", "D", "E", "Gb"]
    assert ["G", "A", "B", "C", "D", "E", "F#"] == scale_to_enharmonic(scale)



import os
from pathlib import Path

from fretboardgtr.fretboard import FretBoard, FretBoardConfig
from fretboardgtr.notes_creators import ScaleFromName

C_IONIAN_SVG = Path(__file__).parent / "data" / "c_ionian.svg"
OUTPUTS_ARTIFACT_FOLDER = Path(__file__).parent / "data" / "outputs" / "e2e"


def remove_test_file(path: Path):
    if path.exists():
        os.remove(str(path))


def test_c_major_fretboard():
    fretboard = FretBoard()
    c_major = ScaleFromName(root="C", mode="Ionian").get()
    fretboard.add_notes(scale=c_major)
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "horizontal.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_c_major_vertical_fretboard():
    fretboard = FretBoard(vertical=True)
    c_major = ScaleFromName(root="C", mode="Ionian").get()
    fretboard.add_notes(scale=c_major)
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "vertical.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_c_major_fretboard_starting_at_5():
    config = {
        "general": {
            "first_fret": 5,
            "last_fret": 12,
        },
    }

    fretboard_config = FretBoardConfig.from_dict(config)
    fretboard = FretBoard(config=fretboard_config)
    c_major = ScaleFromName(root="C", mode="Ionian").get()
    fretboard.add_notes(scale=c_major)
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "horizontal_starting_at_5.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_c_major_vertical_fretboard_starting_at_5():
    config = {
        "general": {
            "first_fret": 5,
            "last_fret": 12,
        },
    }

    fretboard_config = FretBoardConfig.from_dict(config)
    fretboard = FretBoard(config=fretboard_config, vertical=True)
    c_major = ScaleFromName(root="C", mode="Ionian").get()
    fretboard.add_notes(scale=c_major)
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "vertical_starting_at_5.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_c_major_fingering():
    config = {
        "general": {
            "first_fret": 0,
            "last_fret": 5,
            "fret_width": 50,
        }
    }
    fretboard_config = FretBoardConfig.from_dict(config)
    fretboard = FretBoard(config=fretboard_config)
    c_major = [0, 3, 2, 0, 1, 0]

    fretboard.add_fingering(c_major, root="C")
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "c_major_fingering.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_c_major_vertical_fingering():
    config = {
        "general": {
            "first_fret": 0,
            "last_fret": 5,
            "fret_width": 50,
        }
    }
    fretboard_config = FretBoardConfig.from_dict(config)
    fretboard = FretBoard(config=fretboard_config, vertical=True)
    c_major = [0, 3, 2, 0, 1, 0]

    fretboard.add_fingering(c_major, root="C")
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "c_major_vertical_fingering.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_a_major_pentatonic_scale():
    config = {
        "general": {
            "first_fret": 0,
            "last_fret": 12,
            "fret_width": 50,
        }
    }
    fretboard_config = FretBoardConfig.from_dict(config)
    fretboard = FretBoard(config=fretboard_config)
    a_major_pentatonic = [[5, 8], [5, 7], [5, 7], [5, 7], [5, 8], [5, 8]]

    fretboard.add_scale(a_major_pentatonic, root="A")
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "a_major_pentatonic.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_a_major_vertical_pentatonic_scale():
    config = {
        "general": {
            "first_fret": 0,
            "last_fret": 12,
            "fret_width": 50,
        }
    }
    fretboard_config = FretBoardConfig.from_dict(config)
    fretboard = FretBoard(config=fretboard_config, vertical=True)
    a_major_pentatonic = [[5, 8], [5, 7], [5, 7], [5, 7], [5, 8], [5, 8]]

    fretboard.add_scale(a_major_pentatonic, root="A")
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "a_major_pentatonic_vertical.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()



from fretboardgtr.elements.fret_number import FretNumber, FretNumberConfig


def test_fret_number_get_svg():
    fret_number = FretNumber("test", position=(0.0, 0.0))
    attribs = fret_number.get_svg().attribs
    assert float(attribs["x"]) == 0.0
    assert float(attribs["y"]) == 0.0


def test_fret_number_get_svg_custom_config():
    fret_number_config = FretNumberConfig(fontsize=30)
    fret_number = FretNumber("test", position=(0.0, 0.0), config=fret_number_config)
    attribs = fret_number.get_svg().attribs
    assert fret_number.get_svg().text == "test"
    assert float(attribs["x"]) == 0.0
    assert float(attribs["y"]) == 0.0
    assert attribs["font-size"] == 30



from fretboardgtr.elements.notes import (
    FrettedNote,
    FrettedNoteConfig,
    OpenNote,
    OpenNoteConfig,
)


def test_open_note_get_svg():
    open_note = OpenNote(name="test", position=(0.0, 0.0))
    circle = open_note.get_svg().elements[0]
    text = open_note.get_svg().elements[1]
    circle_attribs = circle.attribs
    assert circle_attribs["cx"] == 0.0
    assert circle_attribs["cy"] == 0.0

    assert text.text == "test"
    text_attribs = text.attribs
    assert float(text_attribs["x"]) == 0.0
    assert float(text_attribs["y"]) == 0.0


def test_open_note_get_svg_custom_config():
    open_note_config = OpenNoteConfig(radius=30)
    open_note = OpenNote(name="test", position=(0.0, 0.0), config=open_note_config)
    circle = open_note.get_svg().elements[0]
    text = open_note.get_svg().elements[1]
    circle_attribs = circle.attribs
    assert circle_attribs["cx"] == 0.0
    assert circle_attribs["cy"] == 0.0
    assert circle_attribs["r"] == 30

    assert text.text == "test"
    text_attribs = text.attribs
    assert float(text_attribs["x"]) == 0.0
    assert float(text_attribs["y"]) == 0.0


def test_fretted_note_get_svg():
    fretted_note = FrettedNote(name="test", position=(0.0, 0.0))
    circle = fretted_note.get_svg().elements[0]
    text = fretted_note.get_svg().elements[1]
    circle_attribs = circle.attribs
    assert circle_attribs["cx"] == 0.0
    assert circle_attribs["cy"] == 0.0

    assert text.text == "test"
    text_attribs = text.attribs
    assert float(text_attribs["x"]) == 0.0
    assert float(text_attribs["y"]) == 0.0


def test_fretted_note_get_svg_custom_config():
    fretted_note_config = FrettedNoteConfig(radius=30)
    fretted_note = FrettedNote(
        name="test", position=(0.0, 0.0), config=fretted_note_config
    )
    circle = fretted_note.get_svg().elements[0]
    text = fretted_note.get_svg().elements[1]
    circle_attribs = circle.attribs
    assert circle_attribs["cx"] == 0.0
    assert circle_attribs["cy"] == 0.0
    assert circle_attribs["r"] == 30

    assert text.text == "test"
    text_attribs = text.attribs
    assert float(text_attribs["x"]) == 0.0
    assert float(text_attribs["y"]) == 0.0



from dataclasses import dataclass
from typing import Optional, Tuple

import svgwrite

from fretboardgtr.base import ConfigIniter
from fretboardgtr.constants import BLACK
from fretboardgtr.elements.base import FretBoardElement

SVG_OVERLAY = 10  # overlay


@dataclass
class NutConfig(ConfigIniter):
    """Nut element configuration."""

    color: str = BLACK
    width: int = 6


class Nut(FretBoardElement):
    """Nut element to be drawn in the final fretboard."""

    def __init__(
        self,
        start_position: Tuple[float, float],
        end_position: Tuple[float, float],
        config: Optional[NutConfig] = None,
    ):
        self.config = config if config else NutConfig()
        self.start_position = start_position
        self.end_position = end_position

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the Nut to a svgwrite object.

        This maps the NutConfig configuration attributes to the svg
        attributes
        """
        line = svgwrite.shapes.Line(
            start=self.start_position,
            end=self.end_position,
            stroke=self.config.color,
            stroke_width=self.config.width,
        )
        return line



from dataclasses import dataclass
from typing import Optional, Tuple

import svgwrite

from fretboardgtr.base import ConfigIniter
from fretboardgtr.constants import BLACK, DARK_GRAY
from fretboardgtr.elements.base import FretBoardElement


@dataclass
class NeckDotConfig(ConfigIniter):
    """NeckDot element configuration."""

    color: str = DARK_GRAY
    color_stroke: str = BLACK
    width_stroke: int = 2
    radius: int = 7


class NeckDot(FretBoardElement):
    """Neck dots elements to be drawn in the final fretboard."""

    def __init__(
        self, position: Tuple[float, float], config: Optional[NeckDotConfig] = None
    ):
        self.config = config if config else NeckDotConfig()
        self.x = position[0]
        self.y = position[1]

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the NeckDot to a svgwrite object.

        This maps the NeckDotConfig configuration attributes to the svg
        attributes
        """
        circle = svgwrite.shapes.Circle(
            (self.x, self.y),
            r=self.config.radius,
            fill=self.config.color,
            stroke=self.config.color_stroke,
            stroke_width=self.config.width_stroke,
        )
        return circle



from dataclasses import dataclass, field

from fretboardgtr.base import ConfigIniter
from fretboardgtr.elements.background import BackgroundConfig
from fretboardgtr.elements.cross import CrossConfig
from fretboardgtr.elements.fret_number import FretNumberConfig
from fretboardgtr.elements.frets import FretConfig
from fretboardgtr.elements.neck_dots import NeckDotConfig
from fretboardgtr.elements.notes import FrettedNoteConfig, OpenNoteConfig
from fretboardgtr.elements.nut import NutConfig
from fretboardgtr.elements.strings import StringConfig
from fretboardgtr.elements.tuning import TuningConfig
from fretboardgtr.note_colors import NoteColors


@dataclass
class FretBoardGeneralConfig(ConfigIniter):
    """General configuration for a Fretboard."""

    x_start: float = 30.0
    y_start: float = 30.0
    x_end_offset: float = 0.0
    y_end_offset: float = 0.0
    fret_height: int = 50
    fret_width: int = 70
    first_fret: int = 0
    last_fret: int = 12
    show_tuning: bool = True
    show_frets: bool = True
    show_nut: bool = True
    show_degree_name: bool = False
    show_note_name: bool = True
    open_color_scale: bool = False
    fretted_color_scale: bool = True
    open_colors: NoteColors = field(default_factory=NoteColors)
    fretted_colors: NoteColors = field(default_factory=NoteColors)
    enharmonic: bool = True


@dataclass
class FretBoardConfig(ConfigIniter):
    """Configuration for the fretboard but also for all its elements.

    Inside configurations are :

    FretBoardGeneralConfig
    BackgroundConfig
    FretNumberConfig
    NeckDotConfig
    FretConfig
    NutConfig
    TuningConfig
    StringConfig
    OpenNoteConfig
    FrettedNoteConfig
    CrossConfig
    """

    general: FretBoardGeneralConfig = field(default_factory=FretBoardGeneralConfig)
    background: BackgroundConfig = field(default_factory=BackgroundConfig)
    fret_numbers: FretNumberConfig = field(default_factory=FretNumberConfig)
    neck_dots: NeckDotConfig = field(default_factory=NeckDotConfig)
    frets: FretConfig = field(default_factory=FretConfig)
    nut: NutConfig = field(default_factory=NutConfig)
    tuning: TuningConfig = field(default_factory=TuningConfig)
    strings: StringConfig = field(default_factory=StringConfig)
    open_notes: OpenNoteConfig = field(default_factory=OpenNoteConfig)
    fretted_notes: FrettedNoteConfig = field(default_factory=FrettedNoteConfig)
    cross: CrossConfig = field(default_factory=CrossConfig)



from abc import ABC, abstractmethod

import svgwrite


class FretBoardElement(ABC):
    """Interface to implement to define a FretBoard element.

    This simply consists of converting the element to a svgwrite element
    """

    @abstractmethod
    def get_svg(self) -> svgwrite.base.BaseElement:
        pass
